import javax.jms.Destination;
import javax.jms.JMSException;
import javax.jms.MessageConsumer;
import javax.jms.Queue;
import javax.jms.Session;
import javax.jms.Topic;

class SessionHolder {

    public final Session session;
    private final java.util.Map<MessageConsumerKey,MessageConsumer> consumers =
          new java.util.HashMap<MessageConsumerKey,MessageConsumer>();

    public SessionHolder(final Session session) {
        if (session == null) {
            throw new NullPointerException("Session cannot be null!");
        }
        this.session = session;
    }

    public MessageConsumer createCachedConsumer(final Destination destination,
            final String selector, final boolean noLocal) throws JMSException {
        final MessageConsumerKey key = new MessageConsumerKey(destination, selector, noLocal);
        MessageConsumer consumer;
        synchronized(consumers) {
            consumer = consumers.get(key);
            if (consumer == null) {
                consumer = session.createConsumer(destination, selector, noLocal);
                if (consumer != null) consumers.put(key, consumer);
            }
        }
        if (consumer == null) {
            throw new JMSException("Session created null consumer with destination="+
                    destination+", selector="+selector+", noLocal="+noLocal);
        }
        return consumer;
    }

    private static class MessageConsumerKey {

        private final String selector, queueName, topicName;
        private final boolean noLocal;
        private final int hashCode;

        public MessageConsumerKey(final Destination destination,
                final String selector, final boolean noLocal) {
            if (destination == null) {
                throw new NullPointerException("Destination cannot be null!");
            }
            this.selector = selector == null ? "" : selector;
            this.noLocal = noLocal;
            String queue = null;
            if (destination instanceof Queue) {
                try {
                    queue = ((Queue)destination).getQueueName();
                } catch (JMSException ignored) {}
            }
            this.queueName = queue;
            String topic = null;
            if (destination instanceof Topic) {
                try {
                    topic = ((Topic)destination).getTopicName();
                } catch (JMSException ignored) {}
            }
            this.topicName = topic;
            if (this.queueName == null && this.topicName == null) {
                throw new IllegalArgumentException("Invalid Destination "+
                        destination.getClass());
            }
            int hashQueue = 1;
            if (this.queueName != null) {
                hashQueue = this.queueName.hashCode();
            }
            int hashTopic = 2;
            if (this.topicName != null) {
                hashTopic = this.topicName.hashCode();
            }
            this.hashCode = this.selector.hashCode() + hashQueue + hashTopic +
                    (this.noLocal ? 3 : 0);
        }

        @Override
        public int hashCode() {
            return this.hashCode;
        }

        @Override
        public boolean equals(final Object o) {
            if (o == this) {
                return true;
            }
            if (!(o instanceof MessageConsumerKey)) {
                return false;
            }
            final MessageConsumerKey other = (MessageConsumerKey) o;
            if (noLocal != other.noLocal || !selector.equals(other.selector)) {
                return false;
            }
            if (queueName == null) {
                return other.queueName == null &&
                    topicName.equals(other.topicName);
            }
            if (topicName == null) {
                return other.topicName == null &&
                    queueName.equals(other.queueName);
            }
            return queueName.equals(other.queueName) &&
                   topicName.equals(other.topicName);
        }
    }
}
