import java.util.Collection;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Map;
import javax.jms.*;

class SharedSession implements Session {

    private final SessionHolder holder;
    private final Pool<SessionHolder> pool;
    private final Map<String,TemporaryQueue> tmpQueues;
    private final Map<String,TemporaryTopic> tmpTopics;
    private final Collection<MessageProducer> producers = new LinkedList<MessageProducer>();
    private final Collection<MessageConsumer> consumers = new LinkedList<MessageConsumer>();
    private static final String DEFAULT_SELECTOR = null;
    private static final boolean DEFAULT_NO_LOCAL = false;
    private boolean closed;

    SharedSession(final SessionHolder holder, final Pool<SessionHolder> pool,
            final Map<String,TemporaryQueue> tmpQueues,
            final Map<String,TemporaryTopic> tmpTopics) throws JMSException {
        this.holder = holder;
        this.pool = pool;
        this.tmpQueues = tmpQueues;
        this.tmpTopics = tmpTopics;
    }

    private void checkClosed() throws javax.jms.IllegalStateException {
        assert Thread.holdsLock(this);
        if (closed) throw new javax.jms.IllegalStateException("Session has been closed");
    }

    private Object consumersLock() throws JMSException {
        // to prevent deadlock
        if (Thread.holdsLock(this) && !Thread.holdsLock(consumers)) {
            throw new JMSException(
                      "cannot call this method while holding a lock on this JMS session");
        }
        return consumers;
    }

    public synchronized BytesMessage createBytesMessage() throws JMSException {
        checkClosed();
        return holder.session.createBytesMessage();
    }

    public synchronized MapMessage createMapMessage() throws JMSException {
        checkClosed();
        return holder.session.createMapMessage();
    }

    public synchronized Message createMessage() throws JMSException {
        checkClosed();
        return holder.session.createMessage();
    }

    public synchronized ObjectMessage createObjectMessage() throws JMSException {
        checkClosed();
        return holder.session.createObjectMessage();
    }

    public synchronized ObjectMessage createObjectMessage(final java.io.Serializable object)
            throws JMSException {
        checkClosed();
        return holder.session.createObjectMessage(object);
    }

    public synchronized StreamMessage createStreamMessage() throws JMSException {
        checkClosed();
        return holder.session.createStreamMessage();
    }

    public synchronized TextMessage createTextMessage() throws JMSException {
        checkClosed();
        return holder.session.createTextMessage();
    }

    public synchronized TextMessage createTextMessage(final String text) throws JMSException {
        checkClosed();
        return holder.session.createTextMessage(text);
    }

    public synchronized boolean getTransacted() throws JMSException {
        checkClosed();
        return holder.session.getTransacted();
    }

    public synchronized int getAcknowledgeMode() throws JMSException {
        checkClosed();
        return holder.session.getAcknowledgeMode();
    }

    public synchronized void commit() throws JMSException {
        checkClosed();
        holder.session.commit();
    }

    public synchronized void rollback() throws JMSException {
        checkClosed();
        holder.session.rollback();
    }

    public void close() throws JMSException {
        boolean done = false;
        try {
            synchronized(consumersLock()) {
            /*
             * No synchronized(this) is here yet to prevent a deadlock,
             * because calling MessageConsumer#close() could wait for its
             * MessageListener#onMessage(Message) to complete and beware:
             * the code in onMessage could call any synchronized method
             * of this Session!
             */
                try {
                    if (!consumers.isEmpty()) {
                        try {
                            closeNextConsumer(consumers.iterator());
                        } finally {
                            consumers.clear();
                        }
                    }
                } finally {
                    synchronized(this) {
                        if (closed) {
                            done = true;
                            return;
                        }
                        closed = true;
                        if (!producers.isEmpty()) {
                            try {
                                closeNextProducer(producers.iterator());
                            } finally {
                                producers.clear();
                            }
                        }
                    }
                }
            }
            done = true;
        } finally {
            if (!done) {
                // an exception has occurred
                try {
                    holder.session.close();
                } catch (Throwable ignore) {
                    // let the previous exception be thrown
                }
            }
        }
        assert done;
        if (holder.session.getTransacted()) {
            holder.session.rollback();
        } else {
            holder.session.recover();
        }
        if (pool != null) {
            pool.store(holder);
        }
    }

    private static void closeNextProducer(final Iterator<MessageProducer> it) throws JMSException {
        if (it.hasNext()) {
            try {
                it.next().close();
            } finally {
                closeNextProducer(it);
            }
        }
    }

    private static void closeNextConsumer(final Iterator<MessageConsumer> it) throws JMSException {
        if (it.hasNext()) {
            try {
                it.next().close();
            } finally {
                closeNextConsumer(it);
            }
        }
    }

    public synchronized void recover() throws JMSException {
        checkClosed();
        holder.session.recover();
    }

    public MessageListener getMessageListener() throws JMSException {
        throw new UnsupportedOperationException();
    }

    public void setMessageListener(final MessageListener listener) throws JMSException {
        throw new UnsupportedOperationException();
    }

    public void run() {
        throw new UnsupportedOperationException();
    }

    public synchronized MessageProducer createProducer(final Destination destination) throws JMSException {
        checkClosed();
        final MessageProducer producer = holder.session.createProducer(destination);
        if (producer == null) {
            throw new JMSException("Session created null producer for destination "+destination);
        }
        producers.add(producer);
        return producer;
    }

    public MessageConsumer createConsumer(final Destination destination) throws JMSException {
        return createConsumer(destination, DEFAULT_SELECTOR, DEFAULT_NO_LOCAL);
    }

    public MessageConsumer createConsumer(final Destination destination, final String selector)
            throws JMSException {
        return createConsumer(destination, selector, DEFAULT_NO_LOCAL);
    }

    public MessageConsumer createConsumer(final Destination destination, final String selector,
            final boolean noLocal) throws JMSException {
        final MessageConsumer consumer;
        synchronized(consumersLock()) {
            synchronized(this) {
                checkClosed();
                if (destination instanceof WrappedDestination || isTemporaryQueue(destination) ||
                                                                 isTemporaryTopic(destination)) {
                    Destination realDestination = destination;
                    while (realDestination instanceof WrappedDestination) {
                        realDestination = ((WrappedDestination)realDestination).destination;
                    }
                    consumer = holder.session.createConsumer(realDestination, selector, noLocal);
                    if (consumer == null) {
                        throw new JMSException("Session created null consumer with destination="+
                                realDestination+", selector="+selector+", noLocal="+noLocal);
                    }
                } else {
                    consumer = new SharedMessageConsumer(holder.createCachedConsumer(destination, selector, noLocal));
                }
            }
            consumers.add(consumer);
        }
        return consumer;
    }

    private boolean isTemporaryQueue(final Destination destination) {
        if (destination instanceof TemporaryQueue) {
            try {
                final String name = ((TemporaryQueue)destination).getQueueName();
                if (name != null) {
                    return tmpQueues.containsKey(name);
                }
            } catch (JMSException ignore) {}
        }
        return false;
    }

    private boolean isTemporaryTopic(final Destination destination) {
        if (destination instanceof TemporaryTopic) {
            try {
                final String name = ((TemporaryTopic)destination).getTopicName();
                if (name != null) {
                    return tmpTopics.containsKey(name);
                }
            } catch (JMSException ignore) {}
        }
        return false;
    }

    public synchronized Queue createQueue(final String name) throws JMSException {
        checkClosed();
        return holder.session.createQueue(name);
    }

    public synchronized Topic createTopic(final String name) throws JMSException {
        checkClosed();
        return holder.session.createTopic(name);
    }

    public TopicSubscriber createDurableSubscriber(final Topic topic, final String name)
            throws JMSException {
        final TopicSubscriber result;
        synchronized(consumersLock()) {
            synchronized(this) {
                checkClosed();
                result = holder.session.createDurableSubscriber(topic, name);
            }
            if (result == null) {
                throw new JMSException("Session created null durable subscriber with name '"+
                        name+"' for topic "+topic);
            }
            consumers.add(result);
        }
        return result;
    }

    public TopicSubscriber createDurableSubscriber(final Topic topic, final String name,
            final String selector, final boolean noLocal) throws JMSException {
        final TopicSubscriber result;
        synchronized(consumersLock()) {
            synchronized(this) {
                checkClosed();
                result = holder.session.createDurableSubscriber(topic, name, selector, noLocal);
            }
            if (result == null) {
                throw new JMSException("Session created null durable subscriber with name="+
                        name+", selector="+selector+", noLocal="+noLocal+" for topic "+topic);
            }
            consumers.add(result);
        }
        return result;
    }

    public synchronized QueueBrowser createBrowser(final Queue queue) throws JMSException {
        checkClosed();
        return holder.session.createBrowser(queue);
    }

    public synchronized QueueBrowser createBrowser(final Queue queue, final String selector)
            throws JMSException {
        checkClosed();
        return holder.session.createBrowser(queue, selector);
    }

    public synchronized TemporaryQueue createTemporaryQueue() throws JMSException {
        checkClosed();
        final TemporaryQueue tmpQueue = holder.session.createTemporaryQueue();
        if (tmpQueue != null) {
            final String name = tmpQueue.getQueueName();
            if (name == null) {
                throw new JMSException("TemporaryQueue with null name!");
            }
            if (tmpQueues.put(name, tmpQueue) != null) {
                throw new javax.jms.IllegalStateException(
                      "There already was TemporaryQueue with name '"+name+"'!");
            }
        }
        return tmpQueue;
    }

    public synchronized TemporaryTopic createTemporaryTopic() throws JMSException {
        checkClosed();
        final TemporaryTopic tmpTopic = holder.session.createTemporaryTopic();
        if (tmpTopic != null) {
            final String name = tmpTopic.getTopicName();
            if (name == null) {
                throw new JMSException("TemporaryTopic with null name!");
            }
            if (tmpTopics.put(name, tmpTopic) != null) {
                throw new javax.jms.IllegalStateException(
                      "There already was TemporaryTopic with name '"+name+"'!");
            }
        }
        return tmpTopic;
    }

    public synchronized void unsubscribe(final String name) throws JMSException {
        checkClosed();
        holder.session.unsubscribe(name);
    }

}
